@@ -6,7 +6,6 @@ This is the Kodexa CLI, it can be used to allow you to work with an instance of
 
 It supports interacting with the API, listing and viewing components.  Note it can also be used to login and logout
 """
-import glob
 import json
 import logging
 import os
@@ -14,15 +13,16 @@ import os.path
 import sys
 import tarfile
 from getpass import getpass
-from pathlib import Path
 from typing import Optional
 
 import click
 import yaml
 from rich import print
 
+logging.root.addHandler(logging.StreamHandler(sys.stdout))
+
+from kodexa import KodexaClient
 from kodexa.cli.documentation import generate_site
-from kodexa.model.model import ModelContentMetadata
 from kodexa.platform.kodexa import ExtensionHelper, KodexaPlatform
 
 LOGGING_LEVELS = {
@@ -47,131 +47,58 @@ class Info(object):
 pass_info = click.make_pass_decorator(Info, ensure=True)
 
 
+def merge(a, b, path=None):
+    "merges dictionary b into dictionary a"
+    if path is None: path = []
+    for key in b:
+        if key in a:
+            if isinstance(a[key], dict) and isinstance(b[key], dict):
+                merge(a[key], b[key], path + [str(key)])
+            elif a[key] == b[key]:
+                pass  # same leaf value
+            else:
+                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
+        else:
+            a[key] = b[key]
+    return a
+
+
 # Change the options to below to suit the actual options for your task (or
 # tasks).
 @click.group()
 @click.option("--verbose", "-v", count=True, help="Enable verbose output.")
 @pass_info
 def cli(info: Info, verbose: int):
-    """Run Kodexa.
-
-    Args:
-      info: Info: 
-      verbose: int: 
-
-    Returns:
-
-    """
     # Use the verbosity count to determine the logging level...
     if verbose > 0:
-        logging.basicConfig(
-            level=LOGGING_LEVELS[verbose]
+        logging.root.setLevel(
+            LOGGING_LEVELS[verbose]
             if verbose in LOGGING_LEVELS
             else logging.DEBUG
         )
         click.echo(
             click.style(
                 f"Verbose logging is enabled. "
-                f"(LEVEL={logging.getLogger().getEffectiveLevel()})",
+                f"(LEVEL={logging.root.getEffectiveLevel()})",
                 fg="yellow",
             )
         )
     info.verbose = verbose
 
 
-@click.option('--path', default=os.getcwd(), help='Path to folder containing kodexa.yml')
-@click.option('--url', default=KodexaPlatform.get_url(), help='The URL to the Kodexa server')
-@click.option('--org', help='The slug for the organization to deploy to', required=True)
-@click.option('--slug', help='The slug to deploy the model to', required=False)
-@click.option('--version', help='The version to deploy the model to', required=False)
-@click.option('--token', default=KodexaPlatform.get_access_token(), help='Access token')
-@cli.command()
-@pass_info
-def push_model(_: Info, path: str, url: str, org: str, token: str, slug: str, version: str):
-    """Deploy extension pack to a Kodexa platform instance
-    """
-
-    print("Pushing model from path", path)
-    KodexaPlatform.set_url(url)
-    KodexaPlatform.set_access_token(token)
-
-    # We need to open the model.yml - upload the contents
-    # and then update the metadata
-    with open(path + "/model.yml", "r") as model_meta_file:
-        import yaml
-        model_meta = yaml.safe_load(model_meta_file)
-
-    slug = slug if slug else model_meta["slug"]
-    version = version if version else model_meta["version"]
-    model_meta["type"] = "model"
-
-    model_meta["version"] = version
-    model_meta["slug"] = slug
-
-    from kodexa import RemoteModelStore
-    ref = f"{org}/{slug}:{version}"
-    print("Pushing model to ", ref)
-    model_meta['ref'] = ref
-    remote_model_store = RemoteModelStore.parse_obj(model_meta)
-
-    KodexaPlatform.deploy(ref, remote_model_store, name=model_meta['name'], description=model_meta['description'],
-                          force_replace=True)
-
-    print("Deleting existing contents")
-    for path in remote_model_store.list_contents():
-        remote_model_store.delete(path)
-
-    for path in model_meta["contents"]:
-        for path_hit in glob.glob(path):
-            print(f"Uploading {path_hit}")
-            if Path(path_hit).is_file():
-                with open(path_hit, 'rb') as path_content:
-                    remote_model_store.put(path_hit, path_content, replace=True)
-
-    model_metadata = ModelContentMetadata.parse_obj(model_meta['metadata'])
-    remote_model_store.set_content_metadata(model_metadata)
-    print("Pushed model :tada:")
-
-
-@click.option('--path', default=os.getcwd(), help='Path to folder containing kodexa.yml')
-@click.option('--url', default=KodexaPlatform.get_url(), help='The URL to the Kodexa server')
-@click.option('--org', help='The slug for the organization to deploy to', required=True)
-@click.option('--token', default=KodexaPlatform.get_access_token(), help='Access token')
-@cli.command()
-@pass_info
-def deploy(_: Info, path: str, url: str, org: str, token: str):
-    """Deploy extension pack to a Kodexa platform instance
-    """
-
-    print("Starting deployment from path", path)
-    KodexaPlatform.set_url(url)
-    KodexaPlatform.set_access_token(token)
-
-    if '://' in path:
-        print("Deploying from URI", path)
-        KodexaPlatform.deploy_extension_from_uri(path, org)
-    else:
-        print("Deploying local metadata from", path)
-        metadata = ExtensionHelper.load_metadata(path)
-        metadata['orgSlug'] = org;
-        KodexaPlatform.deploy_extension(metadata)
-
-    print("Deployed extension pack :tada:")
-
-
 @cli.command()
 @click.argument('id', required=True)
 @click.option('--url', default=KodexaPlatform.get_url(), help='The URL to the Kodexa server')
 @click.option('--token', default=KodexaPlatform.get_access_token(), help='Access token')
 @pass_info
-def project(_: Info, id: str, token: str, url: str):
+def project(_: Info, project_id: str, token: str, url: str):
     """Get all the details for a specific project
     """
 
     KodexaPlatform.set_url(url)
     KodexaPlatform.set_access_token(token)
 
-    KodexaPlatform.get_project(id)
+    KodexaPlatform.get_project(project_id)
 
 
 @cli.command()
@@ -194,15 +121,24 @@ def upload(_: Info, ref: str, path: str, token: str, url: str):
 
 @cli.command()
 @click.option('--org', help='The slug for the organization to deploy to', required=False)
+@click.option('--slug', help='Override the slug for component (only works for a single component)', required=False)
+@click.option('--version', help='Override the version for component (only works for a single component)',
+              required=False)
 @click.option('--file', help='The path to the file containing the object to apply')
+@click.option('--update/--no-update', help='The path to the file containing the object to apply',
+              default=False)
+@click.option('--url', default=KodexaPlatform.get_url(), help='The URL to the Kodexa server')
 @click.option('--token', default=KodexaPlatform.get_access_token(), help='Access token')
 @click.option('--format', default=None, help='The format to input if from stdin (json, yaml)')
+@click.option('--overlay', default=None, help='A JSON or YAML file that will overlay the metadata')
 @pass_info
-def apply(_: Info, org: Optional[str], file: str, token: str, format=None):
-    """Apply an object to a Kodexa platform instance from a file
+def deploy(_: Info, org: Optional[str], file: str, url: str, token: str, format=None, update: bool = False,
+           version=None, overlay: Optional[str] = None, slug=None):
+    """
+    Deploy a component to a Kodexa platform instance from a file or stdin
     """
 
-    KodexaPlatform.set_access_token(token)
+    client = KodexaClient(access_token=token, url=url)
 
     obj = None
     if file is None:
@@ -223,15 +159,48 @@ def apply(_: Info, org: Optional[str], file: str, token: str, format=None):
             else:
                 raise Exception("Unsupported file type")
 
+    overlay_obj = None
+
+    if overlay is not None:
+        print("Reading overlay")
+        if overlay.endswith('yaml') or overlay.endswith('yml'):
+            overlay_obj = yaml.safe_load(sys.stdin.read())
+        elif overlay.endswith('json'):
+            overlay_obj = json.loads(sys.stdin.read())
+        else:
+            raise Exception("Unable to determine the format of the overlay file, must be .json or .yml/.yaml")
+
     if isinstance(obj, list):
         print(f"Found {len(obj)} components")
         for o in obj:
-            print("Applying component")
-            KodexaPlatform.apply(o, org)
+
+            if overlay_obj:
+                o = merge(o, overlay_obj)
+
+            component = client.deserialize(o)
+            if org is not None:
+                component.org_slug = org
+            print(f"Deploying component {component.slug}:{component.version}")
+            component.deploy(update=update)
+
     else:
-        print("Applying component")
-        KodexaPlatform.apply(obj, org)
-    print("Applied component :tada:")
+
+        if overlay_obj:
+            obj = merge(obj, overlay_obj)
+
+        component = client.deserialize(obj)
+
+        if version is not None:
+            component.version = version
+        if slug is not None:
+            component.slug = slug
+        if org is not None:
+            component.org_slug = org
+        print(f"Deploying component {component.slug}:{component.version}")
+        log_details = component.deploy(update=update)
+        for log_detail in log_details:
+            print(log_detail)
+    print("Deployed :tada:")
 
 
 @cli.command()
@@ -274,18 +243,50 @@ def get(_: Info, object_type: str, ref: Optional[str], url: str, token: str, que
 @click.argument('query', default="*")
 @click.option('--url', default=KodexaPlatform.get_url(), help='The URL to the Kodexa server')
 @click.option('--token', default=KodexaPlatform.get_access_token(), help='Access token')
-@click.option('--download/--no-download', default=False, help='Download the KDDB for the lastest in the family')
+@click.option('--download/--no-download', default=False, help='Download the KDDB for the latest in the family')
+@click.option('--download-native/--no-download-native', default=False, help='Download the native file for the family')
 @click.option('--page', default=1, help='Page number')
 @click.option('--pageSize', default=10, help='Page size')
 @click.option('--sort', default=None, help='Sort by ie. name:asc')
 @pass_info
-def query(_: Info, query: str, ref: str, url: str, token: str, download: bool, page: int, pagesize: int, sort: None):
+def query(_: Info, query: str, ref: str, url: str, token: str, download: bool, download_native: bool, page: int,
+          pagesize: int, sort: None):
     """
     Query the documents in a given document store
     """
     KodexaPlatform.set_url(url)
     KodexaPlatform.set_access_token(token)
-    KodexaPlatform.query(ref, query, download, page, pagesize, sort)
+    KodexaPlatform.query(ref, query, download, page, pagesize, sort, download_native)
+
+
+@cli.command()
+@click.argument('project_id', required=True)
+@click.option('--url', default=KodexaPlatform.get_url(), help='The URL to the Kodexa server')
+@click.option('--token', default=KodexaPlatform.get_access_token(), help='Access token')
+@click.option('--output', help='The path to export to')
+@pass_info
+def export_project(_: Info, project_id: str, url: str, token: str, output: str):
+    client = KodexaClient(url, token)
+    project_endpoint = client.projects.get(project_id)
+    client.export_project(project_endpoint, output)
+
+
+@cli.command()
+@click.argument('org_slug', required=True)
+@click.argument('path', required=True)
+@click.option('--url', default=KodexaPlatform.get_url(), help='The URL to the Kodexa server')
+@click.option('--token', default=KodexaPlatform.get_access_token(), help='Access token')
+@pass_info
+def import_project(_: Info, org_slug: str, url: str, token: str, path: str):
+    print("Importing project from {}".format(path))
+
+    client = KodexaClient(url, token)
+    organization = client.organizations.find_by_slug(org_slug)
+
+    print("Organization: {}".format(organization.name))
+    client.import_project(organization, path)
+
+    print("Project imported")
 
 
 @cli.command()
@@ -296,7 +297,7 @@ def query(_: Info, query: str, ref: str, url: str, token: str, download: bool, p
 @click.option('--file', help='The path to the file containing the event to send')
 @click.option('--format', default=None, help='The format to use if from stdin (json, yaml)')
 @pass_info
-def send_event(_: Info, project_id: str, assistant_id: str, url: str, file: str, format: str, token: str):
+def send_event(_: Info, project_id: str, assistant_id: str, url: str, file: str, event_format: str, token: str):
     """Send an event to an assistant
     """
 
@@ -306,9 +307,9 @@ def send_event(_: Info, project_id: str, assistant_id: str, url: str, file: str,
     obj = None
     if file is None:
         print("Reading from stdin")
-        if format == 'yaml':
+        if event_format == 'yaml':
             obj = yaml.parse(sys.stdin.read())
-        elif format == 'json':
+        elif event_format == 'json':
             obj = json.loads(sys.stdin.read())
         else:
             raise Exception("You must provide a format if using stdin")
@@ -329,17 +330,18 @@ def send_event(_: Info, project_id: str, assistant_id: str, url: str, file: str,
 
 @cli.command()
 @click.argument('object_type', required=True)
-@click.argument('ref', required=True)
 @click.option('--url', default=KodexaPlatform.get_url(), help='The URL to the Kodexa server')
 @click.option('--token', default=KodexaPlatform.get_access_token(), help='Access token')
 @pass_info
-def reindex(_: Info, object_type: str, ref: str, url: str, token: str):
+def reindex(_: Info, object_type: str, url: str, token: str):
     """
     Reindex the given resource (based on ref)
     """
+    print("Starting global reindexing")
     KodexaPlatform.set_url(url)
     KodexaPlatform.set_access_token(token)
-    KodexaPlatform.reindex(object_type, ref)
+    KodexaPlatform.reindex(object_type)
+    print(":tada: Global reindexing complete")
 
 
 @cli.command()
@@ -428,6 +430,13 @@ def document(_: Info, path: str):
     print("Extension documentation has been successfully built :tada:")
 
 
+@cli.command()
+@pass_info
+def version(_: Info):
+    import pkg_resources
+    print("Kodexa Version:", pkg_resources.get_distribution("kodexa").version)
+
+
 @cli.command()
 @click.option('--path', default=os.getcwd(), help='Path to folder container kodexa.yml (defaults to current)')
 @click.option('--output', default=os.getcwd() + "/dist",
